"""Module containing the COFOG class."""


from dataclasses import dataclass
from typing import List
from typing import Optional

from cofog.data import COFOG_GLOSSARY
from cofog.exceptions import InvalidCOFOGCodeError
from cofog.exceptions import InvalidCOFOGLevelError
from cofog.exceptions import NoChildrenError
from cofog.exceptions import NoParentError
from cofog.parser import get_all_levels
from cofog.parser import get_children_of
from cofog.parser import get_level
from cofog.parser import parse
from cofog.types import COFOG_CODE_TYPE


@dataclass
class COFOG:
    """A class to represent a COFOG.

    Raises:
        InvalidCOFOGCodeError: When an invalid COFOG code is given.
        InvalidCOFOGLevelError: When an invalid COFOG level is \
            specified in the set_level method.
        ValueError: When the specified level is higher than the \
            original level in the set_level method.

    Returns:
        cofog.COFOG: a COFOG class object.
    """

    code: COFOG_CODE_TYPE

    def __post_init__(self) -> None:
        """Method to run after initialisation of the COFOG class."""
        self.code = parse(self.code)
        if not self.is_valid():
            raise InvalidCOFOGCodeError(self.code)
        self.description = self.get_description()
        self.level = get_level(self.code)
        self._all_levels = get_all_levels(self.code)

    def __repr__(self) -> str:
        """Representation of the given COFOG."""
        return f"{self.code}: {self.description}"

    def is_valid(self) -> bool:
        """Determines whether the given COFOG is valid."""
        return self.code in COFOG_GLOSSARY

    def get_description(self) -> str:
        """Gets description of the given code."""
        return COFOG_GLOSSARY[str(self.code)]

    def set_level(self, level: int) -> None:
        """Sets the COFOG to the given level."""
        if not 0 < level < 4:
            raise InvalidCOFOGLevelError(level)
        if self._all_levels[level] is None:
            raise ValueError("Level does not exist for given COFOG code.")
        # Reset class attributes
        self.code = self._all_levels[level]  # type: ignore[assignment]
        self.description = self.get_description()
        self.level = get_level(str(self.code))

    def get_parent_code(self) -> Optional[str]:
        """Get code of the parent to the current COFOG."""
        if self.level == 1:
            raise NoParentError()
        return self._all_levels[self.level - 1]

    def get_children_codes(self) -> List[str]:
        """Get codes of the children of the current COFOG."""
        if self.level == 3:
            raise NoChildrenError()
        return get_children_of(str(self.code))



"""Module containing the COFOG data."""


from typing import Dict


COFOG_GLOSSARY: Dict[str, str] = {
    "01": "General public services",
    "01.1": "Executive and legislative organs, financial and fiscal affairs, external affairs",  # noqa: B950
    "01.1.1": "Executive and legislative organs  (CS)",
    "01.1.2": "Financial and fiscal affairs  (CS)",
    "01.1.3": "External affairs  (CS)",
    "01.2": "Foreign economic aid",
    "01.2.1": "Economic aid to developing countries and countries in transition  (CS)",
    "01.2.2": "Economic aid routed through international organizations  (CS)",
    "01.3": "General services",
    "01.3.1": "General personnel services  (CS)",
    "01.3.2": "Overall planning and statistical services  (CS)",
    "01.3.3": "Other general services  (CS)",
    "01.4": "Basic research",
    "01.4.0": "Basic research  (CS)",
    "01.5": "R&D General public services",
    "01.5.0": "R&D General public services  (CS)",
    "01.6": "General public services n.e.c.",
    "01.6.0": "General public services n.e.c.  (CS)",
    "01.7": "Public debt transactions",
    "01.7.0": "Public debt transactions  (CS)",
    "01.8": "Transfers of a general character between different levels of government",
    "01.8.0": "Transfers of a general character between different levels of government  (CS)",  # noqa: B950
    "02": "Defence",
    "02.1": "Military defence",
    "02.1.0": "Military defence  (CS)",
    "02.2": "Civil defence",
    "02.2.0": "Civil defence  (CS)",
    "02.3": "Foreign military aid",
    "02.3.0": "Foreign military aid  (CS)",
    "02.4": "R&D Defence",
    "02.4.0": "R&D Defence  (CS)",
    "02.5": "Defence n.e.c.",
    "02.5.0": "Defence n.e.c.  (CS)",
    "03": "Public order and safety",
    "03.1": "Police services",
    "03.1.0": "Police services  (CS)",
    "03.2": "Fire-protection services",
    "03.2.0": "Fire-protection services  (CS)",
    "03.3": "Law courts",
    "03.3.0": "Law courts  (CS)",
    "03.4": "Prisons",
    "03.4.0": "Prisons  (CS)",
    "03.5": "R&D Public order and safety",
    "03.5.0": "R&D Public order and safety  (CS)",
    "03.6": "Public order and safety n.e.c.",
    "03.6.0": "Public order and safety n.e.c.  (CS)",
    "04": "Economic affairs",
    "04.1": "General economic, commercial and labour affairs",
    "04.1.1": "General economic and commercial affairs  (CS)",
    "04.1.2": "General labour affairs  (CS)",
    "04.2": "Agriculture, forestry, fishing and hunting",
    "04.2.1": "Agriculture  (CS)",
    "04.2.2": "Forestry  (CS)",
    "04.2.3": "Fishing and hunting  (CS)",
    "04.3": "Fuel and energy",
    "04.3.1": "Coal and other solid mineral fuels  (CS)",
    "04.3.2": "Petroleum and natural gas  (CS)",
    "04.3.3": "Nuclear fuel  (CS)",
    "04.3.4": "Other fuels  (CS)",
    "04.3.5": "Electricity  (CS)",
    "04.3.6": "Non-electric energy  (CS)",
    "04.4": "Mining, manufacturing and construction",
    "04.4.1": "Mining of mineral resources other than mineral fuels  (CS)",
    "04.4.2": "Manufacturing  (CS)",
    "04.4.3": "Construction  (CS)",
    "04.5": "Transport",
    "04.5.1": "Road transport  (CS)",
    "04.5.2": "Water transport  (CS)",
    "04.5.3": "Railway transport  (CS)",
    "04.5.4": "Air transport  (CS)",
    "04.5.5": "Pipeline and other transport  (CS)",
    "04.6": "Communication",
    "04.6.0": "Communication  (CS)",
    "04.7": "Other industries",
    "04.7.1": "Distributive trades, storage and warehousing  (CS)",
    "04.7.2": "Hotels and restaurants  (CS)",
    "04.7.3": "Tourism  (CS)",
    "04.7.4": "Multi-purpose development projects  (CS)",
    "04.8": "R&D Economic affairs",
    "04.8.1": "R&D General economic, commercial and labour affairs  (CS)",
    "04.8.2": "R&D Agriculture, forestry, fishing and hunting  (CS)",
    "04.8.3": "R&D Fuel and energy  (CS)",
    "04.8.4": "R&D Mining, manufacturing and construction  (CS)",
    "04.8.5": "R&D Transport  (CS)",
    "04.8.6": "R&D Communication  (CS)",
    "04.8.7": "R&D Other industries  (CS)",
    "04.9": "Economic affairs n.e.c.",
    "04.9.0": "Economic affairs n.e.c.  (CS)",
    "05": "Environmental protection",
    "05.1": "Waste management",
    "05.1.0": "Waste management  (CS)",
    "05.2": "Waste water management",
    "05.2.0": "Waste water management  (CS)",
    "05.3": "Pollution abatement",
    "05.3.0": "Pollution abatement  (CS)",
    "05.4": "Protection of biodiversity and landscape",
    "05.4.0": "Protection of biodiversity and landscape  (CS)",
    "05.5": "R&D Environmental protection",
    "05.5.0": "R&D Environmental protection  (CS)",
    "05.6": "Environmental protection n.e.c.",
    "05.6.0": "Environmental protection n.e.c.  (CS)",
    "06": "Housing and community amenities",
    "06.1": "Housing development",
    "06.1.0": "Housing development  (CS)",
    "06.2": "Community development",
    "06.2.0": "Community development  (CS)",
    "06.3": "Water supply",
    "06.3.0": "Water supply  (CS)",
    "06.4": "Street lighting",
    "06.4.0": "Street lighting  (CS)",
    "06.5": "R&D Housing and community amenities",
    "06.5.0": "R&D Housing and community amenities  (CS)",
    "06.6": "Housing and community amenities n.e.c.",
    "06.6.0": "Housing and community amenities n.e.c.  (CS)",
    "07": "Health",
    "07.1": "Medical products, appliances and equipment",
    "07.1.1": "Pharmaceutical products  (IS)",
    "07.1.2": "Other medical products  (IS)",
    "07.1.3": "Therapeutic appliances and equipment  (IS)",
    "07.2": "Outpatient services",
    "07.2.1": "General medical services  (IS)",
    "07.2.2": "Specialized medical services  (IS)",
    "07.2.3": "Dental services  (IS)",
    "07.2.4": "Paramedical services  (IS)",
    "07.3": "Hospital services",
    "07.3.1": "General hospital services  (IS)",
    "07.3.2": "Specialized hospital services  (IS)",
    "07.3.3": "Medical and maternity centre services  (IS)",
    "07.3.4": "Nursing and convalescent home services  (IS)",
    "07.4": "Public health services",
    "07.4.0": "Public health services  (IS)",
    "07.5": "R&D Health",
    "07.5.0": "R&D Health  (CS)",
    "07.6": "Health n.e.c.",
    "07.6.0": "Health n.e.c.  (CS)",
    "08": "Recreation, culture and religion",
    "08.1": "Recreational and sporting services",
    "08.1.0": "Recreational and sporting services  (IS)",
    "08.2": "Cultural services",
    "08.2.0": "Cultural services  (IS)",
    "08.3": "Broadcasting and publishing services",
    "08.3.0": "Broadcasting and publishing services  (CS)",
    "08.4": "Religious and other community services",
    "08.4.0": "Religious and other community services  (CS)",
    "08.5": "R&D Recreation, culture and religion",
    "08.5.0": "R&D Recreation, culture and religion  (CS)",
    "08.6": "Recreation, culture and religion n.e.c.",
    "08.6.0": "Recreation, culture and religion n.e.c.  (CS)",
    "09": "Education",
    "09.1": "Pre-primary and primary education",
    "09.1.1": "Pre-primary education  (IS)",
    "09.1.2": "Primary education  (IS)",
    "09.2": "Secondary education",
    "09.2.1": "Lower-secondary education  (IS)",
    "09.2.2": "Upper-secondary education  (IS)",
    "09.3": "Post-secondary non-tertiary education",
    "09.3.0": "Post-secondary non-tertiary education  (IS)",
    "09.4": "Tertiary education",
    "09.4.1": "First stage of tertiary education  (IS)",
    "09.4.2": "Second stage of tertiary education  (IS)",
    "09.5": "Education not definable by level",
    "09.5.0": "Education not definable by level  (IS)",
    "09.6": "Subsidiary services to education",
    "09.6.0": "Subsidiary services to education  (IS)",
    "09.7": "R&D Education",
    "09.7.0": "R&D Education  (CS)",
    "09.8": "Education n.e.c.",
    "09.8.0": "Education n.e.c.  (CS)",
    "10": "Social protection",
    "10.1": "Sickness and disability",
    "10.1.1": "Sickness  (IS)",
    "10.1.2": "Disability  (IS)",
    "10.2": "Old age",
    "10.2.0": "Old age  (IS)",
    "10.3": "Survivors",
    "10.3.0": "Survivors  (IS)",
    "10.4": "Family and children",
    "10.4.0": "Family and children  (IS)",
    "10.5": "Unemployment",
    "10.5.0": "Unemployment  (IS)",
    "10.6": "Housing",
    "10.6.0": "Housing  (IS)",
    "10.7": "Social exclusion n.e.c.",
    "10.7.0": "Social exclusion n.e.c.  (IS)",
    "10.8": "R&D Social protection",
    "10.8.0": "R&D Social protection  (CS)",
    "10.9": "Social protection n.e.c.",
    "10.9.0": "Social protection n.e.c.  (CS)",
}



"""Module containing parser functions for COFOG codes."""


import re
from typing import Dict
from typing import List
from typing import Optional

from cofog.data import COFOG_GLOSSARY
from cofog.exceptions import InvalidCOFOGCodeError
from cofog.types import COFOG_CODE_TYPE


def parse(code: COFOG_CODE_TYPE) -> str:
    """Function to parse the given COFOG code."""
    if not isinstance(code, str) and not isinstance(code, int):
        raise TypeError("COFOG code must be a string or integer.")

    # Handle case where str contains dots.
    if isinstance(code, str) and "." in code:
        if not bool(
            re.search(r"^(0[1-9]{1}|[1-9]{1}|10{1})(\.{1}\d{1}){0,2}$", code)
        ):  # noqa: W605
            raise InvalidCOFOGCodeError(code)
        # Append 0 at beginning if needed.
        if not code.startswith("0") and code[1] == ".":
            return "0" + code
        return code

    # Handle case where code is int or str with no dots.
    # Check code is between 1 and 4 digits, inclusive.
    if not 0 < len(str(code)) < 5:
        raise InvalidCOFOGCodeError(code)
    # Check code is a valid format.
    if not bool(
        re.search(r"^(0[1-9]{1}|[1-9]{1}|10{1})(\d{1}){0,2}$", str(code))
    ):  # noqa: W605
        raise InvalidCOFOGCodeError(code)
    # Parse code starting with 0
    if str(code).startswith("0"):
        return "0" + ".".join(str(code)[1:])
    # Parse code starting with 1
    if str(code).startswith("10"):
        return "1" + ".".join(str(code)[1:])
    # Parse code starting with other numbers
    return "0" + ".".join(str(code))


def get_level(code: str) -> int:
    """Function to get the level from the given COFOG code."""
    return code.count(".") + 1


def convert_to_lower_level(code: str, level: int) -> str:
    """Function to convert the given COFOG code to a lower level."""
    if level > get_level(code):
        raise ValueError("Level cannot be greater than the level of the code.")
    return ".".join(code.split(".")[:level])


def get_all_levels(code: str) -> Dict[int, Optional[str]]:  # noqa: D103
    """Function to get all levels from the given COFOG code."""
    level = get_level(code)
    if level == 1:
        level_one_code = code
        level_two_code = None
        level_three_code = None
    if level == 2:
        level_one_code = convert_to_lower_level(code, 1)
        level_two_code = code
        level_three_code = None
    if level == 3:
        level_one_code = convert_to_lower_level(code, 1)
        level_two_code = convert_to_lower_level(code, 2)
        level_three_code = code
    return {
        1: level_one_code,
        2: level_two_code,
        3: level_three_code,
    }


def filter_codes(code_to_test: str, code_reg_ex: str, level: int) -> bool:
    """Function to indicate if a code matches certain criteria.

    Args:
        code_to_test (str): Code to test the criteria on.
        code_reg_ex (str): RE pattern string to search for in code_to_test.
        level (int): Level of code being searched for.

    Returns:
        bool: Boolean indicator wether code_to_test passes the filter.
    """
    if bool(re.search(code_reg_ex, code_to_test)) and get_level(code_to_test) == level:
        return True
    return False


def get_children_of(parent_code: str) -> List[str]:
    """Get codes of the children of the current COFOG."""
    level = get_level(parent_code)
    child_level = level + 1
    children_codes = {
        code: desc
        for code, desc in COFOG_GLOSSARY.items()
        if filter_codes(code, parent_code, child_level)
    }
    return list(children_codes.keys())



"""Module containing the COFOG exception classes."""


import textwrap
from dataclasses import dataclass
from typing import Union


@dataclass
class InvalidCOFOGCodeError(Exception):
    """Raised when an invalid COFOG code is given."""

    code: Union[str, int]

    def __str__(self) -> str:
        """Returns the string representation of the exception."""
        message = f"""\
            {self.code} is not a valid COFOG code.
            Valid COFOG codes start with numbers 1-10 and have 1 to 3 levels.
            Please see the docs for more information.
            """
        return textwrap.dedent(message)


@dataclass
class InvalidCOFOGLevelError(Exception):
    """Raised when an invalid COFOG level is given."""

    level: Union[str, int]

    def __str__(self) -> str:
        """Returns the string representation of the exception."""
        message = f"""\
            {self.level} is not a valid COFOG level.
            Valid COFOG codes can only have 1, 2, or 3 levels.
            """
        return textwrap.dedent(message)


@dataclass
class NoParentError(Exception):
    """Raised when attempting to retrieve a parent of a level 1 code."""

    def __str__(self) -> str:
        """Returns the string representation of the exception."""
        message = """\
            Level 1 COFOGs do not have a parent.
            """
        return textwrap.dedent(message)


@dataclass
class NoChildrenError(Exception):
    """Raised when attempting to retrieve the children of non-parent code."""

    def __str__(self) -> str:
        """Returns the string representation of the exception."""
        message = """\
            The selected COFOG does not have any children.
            """
        return textwrap.dedent(message)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['cofog']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.0.1']

entry_points = \
{'console_scripts': ['cofog = cofog.__main__:main']}

setup_kwargs = {
    'name': 'cofog',
    'version': '2.0.0',
    'description': 'Cofog',
    'long_description': '# Cofog\n\n[![PyPI](https://img.shields.io/pypi/v/cofog.svg)][pypi_]\n[![Status](https://img.shields.io/pypi/status/cofog.svg)][status]\n[![Python Version](https://img.shields.io/pypi/pyversions/cofog)][python version]\n[![License](https://img.shields.io/pypi/l/cofog)][license]\n\n[![Read the documentation at https://cofog.readthedocs.io/](https://img.shields.io/readthedocs/cofog/latest.svg?label=Read%20the%20Docs)][read the docs]\n[![Tests](https://github.com/oliverjwroberts/cofog/workflows/Tests/badge.svg)][tests]\n[![Codecov](https://codecov.io/gh/oliverjwroberts/cofog/branch/main/graph/badge.svg)][codecov]\n\n[![pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white)][pre-commit]\n[![Black](https://img.shields.io/badge/code%20style-black-000000.svg)][black]\n\n[pypi_]: https://pypi.org/project/cofog/\n[status]: https://pypi.org/project/cofog/\n[python version]: https://pypi.org/project/cofog\n[read the docs]: https://cofog.readthedocs.io/\n[tests]: https://github.com/oliverjwroberts/cofog/actions?workflow=Tests\n[codecov]: https://app.codecov.io/gh/oliverjwroberts/cofog\n[pre-commit]: https://github.com/pre-commit/pre-commit\n[black]: https://github.com/psf/black\n\n## Introduction\n\nClassification of the Functions of Government (COFOG) is a classification defined by the United Nations Statistics Division. Its purpose is to "classify the purpose of transactions such as outlays on final consumption expenditure, intermediate consumption, gross capital formation and capital and current transfers, by general government" (from [home page]).\n\nThis intention of this package is to serve a convenient way of parsing and interfacing with the classifications.\n\n## Data\n\nData was sourced from the UN\'s [classifications on economic statistics page] and parsed into a dictionary that can be found in `src/cofog/data.py`.\n\n## Features\n\n- Provides the `COFOG` class to represent a Classification of Functions of Government.\n- Validate `COFOG` codes through Regular Expressions and presence in the data.\n- Lookup descriptions from codes.\n- Traverse the levels of any given code, with the ability to set a lower level without forgetting the original level.\n\n## Requirements\n\n- Python >=3.8\n- Click >=8.0.1\n\n## Installation\n\nYou can install _Cofog_ via [pip] from [PyPI]:\n\n```console\n$ pip install cofog\n```\n\n## Usage\n\nPlease see the [usage page] in the docs for full details.\n\nGet started by initialising a new `COFOG` object with a code. These can be either specified as strings (with and without dots) or integers.\n\n```python\nfrom cofog import COFOG\n\ncofog = COFOG("04.3.6")\n\n# or any of the following\nCOFOG("0436")\nCOFOG("436")\nCOFOG(436)\n```\n\nYou can then access the code\'s description as well as set it to lower level codes.\n\n```python\nprint(cofog.description)\n# Non-electric energy  (CS)\nprint(cofog.level)\n# 3\n\ncofog.set_level(2)\nprint(cofog.code)\n# 04.3\nprint(cofog.description)\n# Fuel and energy\n```\n\nYou can also get parent and children codes of any valid code.\n\n```python\nprint(cofog.get_parent_code())\n# 04.3\n\nprint(COFOG("07.3").get_children_codes())\n# ["07.3.1", "07.3.2", "07.3.3", "07.3.4"]\n```\n\n## Contributing\n\nContributions are very welcome.\nTo learn more, see the [Contributor Guide].\n\n## License\n\nDistributed under the terms of the [MIT license][license],\n_Cofog_ is free and open source software.\n\n## Issues\n\nIf you encounter any problems,\nplease [file an issue] along with a detailed description.\n\n## Credits\n\nThis project was heavily inspired by [@ellsphillips]\'s [govsic] package.\n\nThis project was generated from [@cjolowicz]\'s [Hypermodern Python Cookiecutter] template.\n\n[home page]: https://unstats.un.org/unsd/classifications/Family/Detail/4\n[classifications on economic statistics page]: https://unstats.un.org/unsd/classifications/Econ\n[pip]: https://pip.pypa.io/\n[pypi]: https://pypi.org/\n[@ellsphillips]: https://github.com/ellsphillips\n[govsic]: https://github.com/ellsphillips/govsic\n[@cjolowicz]: https://github.com/cjolowicz\n[hypermodern python cookiecutter]: https://github.com/cjolowicz/cookiecutter-hypermodern-python\n[file an issue]: https://github.com/oliverjwroberts/cofog/issues\n\n<!-- github-only -->\n\n[usage page]: https://cofog.readthedocs.io/en/latest/usage.html\n[contributor guide]: https://github.com/oliverjwroberts/cofog/blob/main/CONTRIBUTING.md\n[license]: https://github.com/oliverjwroberts/cofog/blob/main/LICENSE\n',
    'author': 'Oliver Roberts',
    'author_email': 'oliverjwroberts@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/oliverjwroberts/cofog',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
